# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import os
from pathlib import Path

from pytest import mark, skip

from lenskit.config import locate_configuration_root

LK_ROOT = Path(__file__).parent.parent.parent


def test_locate_lktoml_cwd(tmp_path: Path):
    lkt = tmp_path / "lenskit.toml"
    lkt.touch()

    path = locate_configuration_root(cwd=tmp_path)
    assert path == tmp_path


def test_locate_lktoml_parent(tmp_path: Path):
    lkt = tmp_path / "lenskit.toml"
    lkt.touch()

    child = tmp_path / "foo" / "bar"
    child.mkdir(parents=True, exist_ok=True)

    path = locate_configuration_root(cwd=child)
    assert path == tmp_path


def test_stop_git(tmp_path: Path):
    lkt = tmp_path / "lenskit.toml"
    lkt.touch()

    (tmp_path / "foo" / ".git").mkdir(parents=True)

    child = tmp_path / "foo" / "bar"
    child.mkdir(parents=True)

    path = locate_configuration_root(cwd=child)
    assert path is None


def test_stop_pyproject(tmp_path: Path):
    lkt = tmp_path / "lenskit.toml"
    lkt.touch()

    child = tmp_path / "foo" / "bar"
    child.mkdir(parents=True)

    (tmp_path / "foo" / "pyproject.toml").touch()

    path = locate_configuration_root(cwd=child)
    assert path is None


def test_locate_none(tmp_path: Path):
    child = tmp_path / "foo" / "bar"
    child.mkdir(parents=True, exist_ok=True)

    path = locate_configuration_root(cwd=child)
    assert path is None



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Tests for the Vocabulary class.
"""

import pickle
from uuid import UUID, uuid4

import numpy as np
import pyarrow as pa

import hypothesis.strategies as st
from hypothesis import assume, given
from pytest import raises, warns

from lenskit.data import Vocabulary
from lenskit.diagnostics import DataWarning


def id_ints():
    li = np.iinfo(np.int64)
    return st.integers(li.min, li.max)


@given(
    st.one_of(
        st.sets(id_ints()),
        st.sets(st.emails()),
    )
)
def test_create_basic(keys: set[int | str]):
    vocab = Vocabulary(keys, reorder=True)
    assert vocab.size == len(keys)
    assert len(vocab) == len(keys)

    index = vocab.index
    assert all(index.values == sorted(keys))


@given(
    st.one_of(
        st.lists(id_ints()),
        st.lists(st.emails()),
    )
)
def test_create_nonunique(keys: list[int | str]):
    uq = set(keys)
    vocab = Vocabulary(keys, reorder=True)
    assert vocab.size == len(uq)
    assert len(vocab) == len(uq)

    index = vocab.index
    assert all(index.values == sorted(uq))


@given(
    st.one_of(
        st.lists(id_ints()),
        st.lists(st.emails()),
    )
)
def test_equal(keys: list[int | str]):
    vocab = Vocabulary(keys, reorder=True)

    v2 = Vocabulary(keys, reorder=True)
    assert v2 == vocab


@given(st.lists(id_ints()), st.sets(id_ints()))
def test_not_equal(keys: list[int], oks: set[int]):
    uq = set(keys)
    assume(oks != uq)

    vocab = Vocabulary(keys, reorder=True)
    v2 = Vocabulary(oks, reorder=True)
    try:
        assert v2 != vocab
    except AssertionError as e:
        e.add_note(f"v1 hash: {vocab._hash}")
        e.add_note(f"v2 hash: {v2._hash}")
        raise e


@given(
    st.one_of(
        st.sets(id_ints()),
        st.sets(st.emails()),
    ),
    st.lists(
        st.one_of(
            id_ints(),
            st.emails(),
        )
    ),
)
def test_contains(keys: set[int] | set[str], qs: set[int | str]):
    vocab = Vocabulary(keys, reorder=True)

    for qk in qs:
        if qk in keys:
            assert qk in vocab
        else:
            assert qk not in vocab


@given(
    st.one_of(
        st.sets(id_ints()),
        st.sets(st.emails()),
    )
)
def test_lookup_id_index(keys: set[int | str]):
    klist = sorted(keys)

    vocab = Vocabulary(keys, reorder=True)
    assert vocab.size == len(klist)
    assert len(vocab) == len(klist)

    # make sure the numbers are right
    assert all([vocab.number(k) == i for (i, k) in enumerate(klist)])

    # make sure the IDs are right
    assert all([vocab.term(i) == k for (i, k) in enumerate(klist)])


@given(
    st.one_of(
        st.sets(id_ints()),
        st.sets(st.emails()),
    ),
    st.one_of(id_ints(), st.emails()),
)
def test_lookup_bad_id(keys: set[int | str], key: int | str):
    assume(key not in keys)

    vocab = Vocabulary(keys, reorder=True)

    assert vocab.number(key, missing=None) is None

    with raises(KeyError):
        assert vocab.number(key, missing="error")


@given(
    st.one_of(
        st.sets(id_ints()),
        st.sets(st.emails()),
    ),
    st.one_of(id_ints()),
)
def test_lookup_bad_number(keys: set[int | str], num: int):
    assume(num < 0 or num >= len(keys))

    vocab = Vocabulary(keys, reorder=True)

    with raises(IndexError):
        vocab.term(num)


@given(
    st.sets(id_ints()),
    st.lists(id_ints()),
)
def test_lookup_many_nums(terms: set[int], lookup: list[int]):
    klist = sorted(terms)
    kpos = dict(zip(klist, range(len(klist))))

    vocab = Vocabulary(terms)

    nums = vocab.numbers(lookup, missing="negative")
    assert len(nums) == len(lookup)
    for n, k in zip(nums, lookup):
        if n < 0:
            assert k not in terms
        else:
            assert n == kpos[k]


@given(
    st.sets(id_ints()),
    st.lists(id_ints()),
)
def test_lookup_many_nums_null(terms: set[int], lookup: list[int]):
    klist = sorted(terms)
    kpos = dict(zip(klist, range(len(klist))))

    vocab = Vocabulary(terms, reorder=True)

    nums = vocab.numbers(lookup, format="arrow", missing="null")
    assert len(nums) == len(lookup)
    assert nums.null_count == len([i for i in lookup if i not in terms])
    for n, k in zip(nums, lookup):
        if n.is_valid:
            assert n.as_py() == kpos[k]
        else:
            assert k not in terms


@given(
    st.sets(st.emails()),
    st.lists(st.emails()),
)
def test_lookup_many_nums_null(terms: set[str], lookup: list[str]):
    klist = sorted(terms)
    kpos = dict(zip(klist, range(len(klist))))

    vocab = Vocabulary(terms, reorder=True)

    nums = vocab.numbers(lookup, format="arrow", missing="null")
    assert len(nums) == len(lookup)
    assert nums.null_count == len([i for i in lookup if i not in terms])
    for n, k in zip(nums, lookup):
        if n.is_valid:
            assert n.as_py() == kpos[k]
        else:
            assert k not in terms


@given(
    st.data(),
    st.sets(id_ints()),
)
def test_lookup_many_terms(data, terms: set[int]):
    assume(len(terms) > 0)
    lookup = data.draw(st.lists(st.integers(min_value=0, max_value=len(terms) - 1)))
    klist = sorted(terms)

    vocab = Vocabulary(terms)

    keys = vocab.terms(lookup)
    assert len(keys) == len(lookup)
    for k, n in zip(keys, lookup):
        assert k == klist[n]


@given(st.one_of(st.sets(id_ints()), st.sets(st.emails())))
def test_all_terms(initial: set[int] | set[str]):
    vocab = Vocabulary(initial)

    tl = sorted(initial)

    terms = vocab.terms()
    assert isinstance(terms, np.ndarray)
    assert all(terms == tl)


def test_lots_of_strings(rng: np.random.Generator):
    "make sure a lot of strings work"
    N = 10_000

    ids = [str(uuid4()) for _i in range(N)]
    vocab = Vocabulary(ids, reorder=True)

    arr = vocab.id_array()
    assert len(arr) == N
    assert arr.null_count == 0

    id_arr = arr.to_numpy(zero_copy_only=False)

    q_pos = rng.choice(N, 5000, replace=True)
    query = id_arr[q_pos]

    nums = vocab.numbers(query, format="arrow", missing="null")
    assert nums.type == pa.int32()
    assert nums.null_count == 0
    assert np.all(nums.to_numpy() >= 0)
    assert np.all(nums.to_numpy() < N)
    assert np.all(id_arr[nums.to_numpy()] == query)


@given(st.one_of(st.lists(id_ints()), st.lists(st.emails()), st.lists(st.emails())))
def test_pickle(initial: list[int | str]):
    vocab = Vocabulary(initial, reorder=True)

    blob = pickle.dumps(vocab)
    v2 = pickle.loads(blob)

    assert v2 is not vocab
    assert len(v2) == len(vocab)
    assert np.all(v2.ids() == vocab.ids())
    assert v2 == vocab



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from __future__ import annotations

import numpy as np
from scipy.sparse import coo_array
from typing_extensions import override

from lenskit._accel import als
from lenskit.basic import BiasModel, Damping
from lenskit.data import Dataset, ItemList
from lenskit.data.types import NPMatrix, NPVector
from lenskit.logging import item_progress
from lenskit.math.solve import solve_cholesky

from ._common import ALSBase, ALSConfig, ALSTrainerBase, TrainContext


class BiasedMFConfig(ALSConfig):
    damping: Damping = 5.0
    """
    Damping for the bias model.
    """


class BiasedMFScorer(ALSBase):
    """
    Biased matrix factorization trained with alternating least squares
    :cite:p:`zhouLargeScaleParallelCollaborative2008`.  This is a
    prediction-oriented algorithm suitable for explicit feedback data, using the
    alternating least squares approach to compute :math:`P` and :math:`Q` to
    minimize the regularized squared reconstruction error of the ratings matrix.

    See the base class :class:`ALSBase` for documentation on the estimated
    parameters you can extract from a trained model. See
    :class:`BiasedMFConfig` and :class:`ALSConfig` for the configuration
    options for this component.

    Stability:
        Caller
    """

    config: BiasedMFConfig
    bias: BiasModel

    def create_trainer(self, data, options):
        return BiasedMFTrainer(self, data, options)

    @override
    def new_user_embedding(
        self, user_num: int | None, items: ItemList
    ) -> tuple[NPVector, float | None]:
        inums = items.numbers(vocabulary=self.items, missing="negative")
        mask = inums >= 0
        ratings = items.field("rating")
        assert ratings is not None
        mask &= np.isfinite(ratings)

        biases, u_bias = self.bias.compute_for_items(items, None, items)
        ratings = ratings - biases

        ri_val = ratings[mask]

        u_feat = _train_bias_row_cholesky(
            inums[mask], ri_val, self.item_embeddings, self.config.user_reg
        )
        return u_feat, u_bias

    @override
    def finalize_scores(
        self, user_num: int | None, items: ItemList, user_bias: float | None
    ) -> ItemList:
        scores = items.scores()
        assert scores is not None

        if user_bias is None:
            if user_num is not None and self.bias.user_biases is not None:
                user_bias = self.bias.user_biases[user_num]
            else:
                user_bias = 0.0
        assert user_bias is not None

        biases = self.bias.compute_for_items(items, bias=user_bias)
        scores = scores + biases

        return ItemList(items, scores=scores)


class BiasedMFTrainer(ALSTrainerBase[BiasedMFScorer, BiasedMFConfig]):
    @override
    def prepare_matrix(self, data: Dataset) -> coo_array:
        # transform ratings using offsets
        rmat = data.interactions().matrix().scipy(attribute="rating", layout="coo")

        self.logger.info("normalizing ratings")
        self.scorer.bias = BiasModel.learn(data, damping=self.config.damping)
        rmat = self.scorer.bias.transform_matrix(rmat)

        return rmat.astype(np.float32)

    @override
    def initial_params(self, nrows: int, ncols: int) -> NPMatrix:
        mat = self.rng.standard_normal((nrows, ncols), dtype=np.float32)
        mat /= np.linalg.norm(mat, axis=1).reshape((nrows, 1))
        return mat

    @override
    def als_half_epoch(self, epoch: int, context: TrainContext) -> float:
        with item_progress(f"Epoch {epoch} {context.label}", context.nrows) as pb:
            return als.train_explicit_matrix(
                context.matrix, context.left, context.right, context.reg, pb
            )


def _train_bias_row_cholesky(
    items: NPVector[np.int32], ratings: NPVector, other: NPMatrix, reg: float
) -> NPVector:
    """
    Args:
        items: the item IDs the user has rated
        ratings: the user's (normalized) ratings for those items
        other: the item-feature matrix
        reg: the regularization term
    Returns:
        the user-feature vector (equivalent to V in the current Cholesky code)
    """
    (nui,) = items.shape
    nf = other.shape[1]

    if nui == 0:
        return np.zeros(nf, dtype=np.float32)

    M = other[items, :]
    regI = np.eye(nf, dtype=np.float32) * reg
    MMT = M.T @ M
    A = MMT + regI * len(items)

    V = M.T @ ratings
    x = solve_cholesky(A, V.astype(A.dtype))

    return np.require(x, dtype=np.float32)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from lenskit.data.schema import CURRENT_VERSION, OLDEST_VERSION, DataSchema


def test_schema_default_version():
    schema = DataSchema(name="foo")
    assert schema.name == "foo"
    assert schema.version == CURRENT_VERSION


def test_schema_load_compat_version():
    schema = DataSchema.model_validate_json("{}")
    assert schema.version == OLDEST_VERSION



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Definition of pipeline hook protocols.
"""

from __future__ import annotations

from collections.abc import Callable

from typing_extensions import (
    Any,
    Generic,
    NamedTuple,
    Protocol,
    TypeAlias,
    TypedDict,
    TypeVar,
)

from ..components import Component, PipelineFunction
from ..nodes import ComponentInstanceNode
from ..types import TypeExpr

ComponentObject: TypeAlias = Component | PipelineFunction

Hook = TypeVar("Hook", bound=Callable[..., Any], covariant=True)


class HookEntry(NamedTuple, Generic[Hook]):
    """
    An entry in a pipeline hook list.
    """

    function: Hook
    priority: int = 1


class ComponentInputHook(Protocol):
    """
    Inspect or process data as it passes to a component's input.

    As with all :ref:`pipeline-hooks`, an input hook is a callable that is run
    at the appropriate stage of the input.

    Component input hooks are installed under the name ``component-input``.
    """

    def __call__(
        self,
        node: ComponentInstanceNode[Any],
        input_name: str,
        input_type: TypeExpr | None,
        value: object,
        **context: Any,
    ) -> Any:
        """
        Inspect or process the component input data.

        Args:
            node:
                The component node being invoked.
            input_name:
                The name of the component's input that will receive the data.
            input_type:
                The type of data the component expects for this input, if one
                was specified in the component definition.
            value:
                The data value to be supplied.  This is declared
                :class:`object`, because its type is not known or guaranteed in
                the genral case.
            context:
                Additional context variables, mostly for the use of internal hooks.

        Returns:
            The value to pass to the component.  For inspection-only hooks, this
            will just be ``value``; hooks can also substitute alternative
            values depending on application needs.
        """
        pass


RunHooks = TypedDict(
    "RunHooks", {"component-input": list[HookEntry[ComponentInputHook]]}, total=False
)


def default_run_hooks() -> RunHooks:
    from .typecheck import typecheck_input_data

    return {"component-input": [HookEntry(typecheck_input_data, 0)]}



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Tests for matrix-vector multiplication to make sure it works properly.

Bug: https://github.com/pytorch/pytorch/issues/127491
"""

import numpy as np
import scipy.sparse as sps
import torch

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import HealthCheck, given, note, settings
from pytest import approx, mark, skip


def draw_problem(
    data, nrows: int, ncols: int, dtype=np.float64
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    # draw the initial matrix
    M = data.draw(
        nph.arrays(
            dtype,
            (nrows, ncols),
            elements=st.floats(
                -1e4,
                1e4,
                allow_nan=False,
                allow_infinity=False,
                width=np.finfo(dtype).bits,  # type: ignore
            ),
        )
    )
    # draw the vector
    v = data.draw(
        nph.arrays(
            dtype,
            ncols,
            elements=st.floats(
                -1e4,
                1e4,
                allow_nan=False,
                allow_infinity=False,
                width=np.finfo(dtype).bits,  # type: ignore
            ),
        )
    )
    # zero out items in the matrix
    mask = data.draw(nph.arrays(np.bool_, (nrows, ncols)))
    M[~mask] = 0.0
    nnz = np.sum(M != 0.0)
    note("matrix {} x {} ({} nnz)".format(nrows, ncols, nnz))

    # multiply them (dense operation with NumPy) to get expected result
    res = M @ v
    # just make sure everything's finite, should always pass due to data limits
    assert np.all(np.isfinite(res))

    return M, v, res


def tolerances(dtype=np.float64):
    # make our tolerance depend on the data type we got
    if dtype == np.float64:
        rtol, atol = 1.0e-5, 1.0e-4
    elif dtype == np.float32:
        rtol, atol = 0.05, 1.0e-3
    else:
        raise TypeError(f"unexpected data type {dtype}")
    return rtol, atol


def torchify(M, v, res, rtol, atol) -> tuple[torch.Tensor, torch.Tensor]:
    T = torch.from_numpy(M)
    tv = torch.from_numpy(v)
    assert torch.mv(T, tv).numpy() == approx(res, rel=rtol, abs=atol)
    return T, tv


def torch_test(func):
    wrapped = given(st.data(), st.integers(1, 500), st.integers(1, 500))(func)
    wrapped = settings(
        deadline=1000, max_examples=500, suppress_health_check=[HealthCheck.too_slow]
    )(wrapped)
    return wrapped


@mark.skip(reason="test unnecessary")
@torch_test
def test_torch_spmv_coo(data, nrows, ncols):
    rtol, atol = tolerances(np.float64)
    M, v, res = draw_problem(data, nrows, ncols)
    T, tv = torchify(M, v, res, rtol, atol)

    TS = T.to_sparse_coo().coalesce()

    tres = torch.mv(TS, tv)
    assert tres.numpy() == approx(res, rel=rtol, abs=atol)


@torch_test
def test_torch_spmv_csr(data, nrows, ncols):
    rtol, atol = tolerances(np.float64)
    M, v, res = draw_problem(data, nrows, ncols)
    T, tv = torchify(M, v, res, rtol, atol)

    TS = T.to_sparse_csr()

    tres = torch.mv(TS, tv)
    assert tres.numpy() == approx(res, rel=rtol, abs=atol)


@mark.skip(
    reason="spmv CSC currently broken",
)
@torch_test
def test_torch_spmv_csc(data, nrows, ncols):
    rtol, atol = tolerances(np.float64)
    M, v, res = draw_problem(data, nrows, ncols)
    T, tv = torchify(M, v, res, rtol, atol)

    TS = T.to_sparse_csc()

    tres = torch.mv(TS, tv)
    assert tres.numpy() == approx(res, rel=rtol, abs=atol)


@mark.skip("not using SPMM fallback")
@torch_test
def test_torch_spmm_coo(data, nrows, ncols):
    rtol, atol = tolerances(np.float64)
    M, v, res = draw_problem(data, nrows, ncols)
    T, tv = torchify(M, v, res, rtol, atol)

    TS = T.to_sparse_coo().coalesce()

    tres = torch.mm(TS, tv.reshape(-1, 1)).reshape(-1)
    assert tres.numpy() == approx(res, rel=rtol, abs=atol)


@mark.skip("not using SPMM fallback")
@torch_test
def test_torch_spmm_csr(data, nrows, ncols):
    rtol, atol = tolerances(np.float64)
    M, v, res = draw_problem(data, nrows, ncols)
    T, tv = torchify(M, v, res, rtol, atol)

    TS = T.to_sparse_csr()

    tres = torch.mm(TS, tv.reshape(-1, 1)).reshape(-1)
    assert tres.numpy() == approx(res, rel=rtol, abs=atol)


@mark.skip("not using SPMM fallback")
@torch_test
def test_torch_spmm_csc(data, nrows, ncols):
    rtol, atol = tolerances(np.float64)
    M, v, res = draw_problem(data, nrows, ncols)
    T, tv = torchify(M, v, res, rtol, atol)

    TS = T.to_sparse_csc()

    tres = torch.mm(TS, tv.reshape(-1, 1)).reshape(-1)
    assert tres.numpy() == approx(res, rel=rtol, abs=atol)


@mark.benchmark(max_time=10)
@mark.parametrize("device", ["cpu", "cuda"])
def test_torch_spmv_bench(device, benchmark):
    if device == "cuda" and not torch.cuda.is_available():
        skip("CUDA disabled")
    A: sps.csr_matrix = sps.random(500, 500, 0.1, format="csr")
    x = np.random.randn(500)

    T = torch.sparse_csr_tensor(
        torch.from_numpy(A.indptr),
        torch.from_numpy(A.indices),
        torch.from_numpy(A.data),
        size=A.shape,
    ).to(device)
    v = torch.from_numpy(x).to(device)

    def op():
        torch.mv(T, v)

    benchmark(op)


@mark.benchmark(max_time=10)
@mark.parametrize("device", ["cpu", "cuda"])
def test_torch_spmm_bench(device, benchmark):
    if device == "cuda" and not torch.cuda.is_available():
        skip("CUDA disabled")
    A: sps.csr_matrix = sps.random(2500, 2500, 0.1, format="csr")
    x = np.random.randn(2500)

    T = torch.sparse_csr_tensor(
        torch.from_numpy(A.indptr),
        torch.from_numpy(A.indices),
        torch.from_numpy(A.data),
        size=A.shape,
    ).to(device)
    v = torch.from_numpy(x).to(device)

    def op():
        torch.mm(T, v.reshape(-1, 1)).reshape(-1)

    benchmark(op)


@mark.benchmark(max_time=10)
def test_scipy_spmv_bench(benchmark):
    A: sps.csr_matrix = sps.random(2500, 2500, 0.1, format="csr")
    x = np.random.randn(2500)

    def op():
        A @ x

    benchmark(op)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Utility code for logging and progress reporting from LensKit components.
"""

from __future__ import annotations

from contextlib import contextmanager
from typing import Generator
from uuid import uuid4

import torch

from ._base import Progress
from ._dispatch import item_progress

_active_pbs: dict[str, Progress] = {}


@contextmanager
def item_progress_handle(
    *args: item_progress.args, **kwargs: item_progress.kwargs
) -> Generator[str, None, None]:
    with item_progress(*args, **kwargs) as pb:
        h = uuid4().hex
        _active_pbs[h] = pb
        try:
            yield h
        finally:
            del _active_pbs[h]


@torch.jit.ignore  # type: ignore
def pbh_update(h, incr=1):
    # type: (str, int) -> None
    # we need old-school annotations to make pytorch happy
    pb = _active_pbs[h]
    pb.update(incr)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Pydantic models for LensKit data schemas.  These models define define the data
schema in memory and also define how schemas are serialized to and from
configuration files.  See :ref:`data-model` for details.

.. note::

    The schema does not specify data types directly — data types are inferred
    from the underlying Arrow data structures.  This reduces duplication of type
    information and the opportunity for inconsistency.
"""

# pyright: strict
from __future__ import annotations

import re
from enum import Enum
from typing import Annotated, Any, Literal, TypeAlias

from pydantic import (
    BaseModel,
    StringConstraints,
    ValidationInfo,
    model_validator,
)

CURRENT_VERSION = "2025.3"
OLDEST_VERSION = "2025.1"

# Pydantic context to use when loading rather than constructing.
LOAD_CONTEXT = {"version": "compat"}

NAME_PATTERN = re.compile(r"^[\w_]+$")
Name: TypeAlias = Annotated[str, StringConstraints(pattern=NAME_PATTERN)]


def id_col_name(name: str) -> str:
    return f"{name}_id"


def num_col_name(name: str) -> str:
    return f"{name}_num"


def check_name(name: str) -> None:
    """
    Check if a name is valid.

    Raises:
        ValueError: when the name is invalid.
    """
    if not NAME_PATTERN.match(name):
        raise ValueError(f"invalid name “{name}”")


class AllowableTroolean(Enum):
    """
    Stores both whether a feature is allowed and is used in the case of having repeats.
    For convenience, in serialized data or configuration files these values may be specified
    either as strings or as booleans, in which case ``False`` is :attr:`FORBIDDEN` and
    ``True`` is :attr:`ALLOWED`. They are always serialized as strings.
    """

    FORBIDDEN = "forbidden"
    """
    The feature is forbidden.
    """
    ALLOWED = "allowed"
    """
    The feature is allowed, but no records using it are present.
    """
    PRESENT = "present"
    """
    The feature is used by instances in the data.
    """

    @property
    def is_allowed(self) -> bool:
        """
        Query whether the feature is allowed.
        """
        return self != self.FORBIDDEN

    @property
    def is_forbidden(self) -> bool:
        """
        Query whether the feature is forbidden.
        """
        return self == self.FORBIDDEN

    @property
    def is_present(self) -> bool:
        """
        Query whether the feature is present (used in recorded instances).
        """
        return self == self.PRESENT

    @model_validator(mode="before")
    @classmethod
    def _validate_policy(cls, value: Any):
        if value is True:
            return AllowableTroolean.ALLOWED
        elif value is False:
            return AllowableTroolean.FORBIDDEN
        else:
            return value


class AttrLayout(Enum):
    SCALAR = "scalar"
    """
    Scalar (non-list, non-vector) attribute value.
    """
    LIST = "list"
    """
    Homogenous, variable-length list of attribute values.
    """
    VECTOR = "vector"
    """
    Homogenous, fixed-length vector of numeric attribute values.
    """
    SPARSE = "sparse"
    """
    Homogenous, fixed-length sparse vector of numeric attribute values.
    """


class DataSchema(BaseModel):
    """
    Description of the entities and layout of a dataset.
    """

    version: str = CURRENT_VERSION
    """
    The data layout version.

    .. note::

        When a new schema model is created, this defaults to the current version
        instead of the oldest version.
    """

    name: str | None = None
    """
    The dataset name.
    """

    default_interaction: Name | None = None
    """
    The default interaction type.
    """

    entities: dict[Name, EntitySchema] = {}
    """
    Entity classes defined for this dataset.
    """
    relationships: dict[Name, RelationshipSchema] = {}
    """
    Relationship classes defined for this dataset.
    """

    @model_validator(mode="before")
    @classmethod
    def _default_version(cls, data: Any, info: ValidationInfo):
        ctx = info.context
        vmode = ctx.get("version", "current") if isinstance(ctx, dict) else None  # type: ignore
        if "version" not in data and vmode == "compat":
            return data | {"version": OLDEST_VERSION}
        else:
            return data

    @classmethod
    def model_validate_json(
        cls, json_data: str | bytes | bytearray, *, context: Any = None, **kwargs: Any
    ):
        if context is None:
            context = LOAD_CONTEXT

        return super().model_validate_json(json_data, context=context, **kwargs)


class EntitySchema(BaseModel):
    """
    Entity class definitions in the dataset schema.
    """

    id_type: Literal["int", "str"] | None = None
    """
    The data type for identifiers in this entity class.
    """
    attributes: dict[Name, ColumnSpec] = {}
    """
    Entity attribute definitions.
    """


class RelationshipSchema(BaseModel):
    """
    Relationship class definitions in the dataset schema.
    """

    entities: dict[Name, Name | None]
    """
    Define the entity classes participating in the relationship.  For aliased
    entity classes (necessary for self-relationships), the key is the alias, and
    the value is the original entity class name.
    """
    interaction: bool = False
    """
    Whether this relationship class records interactions.
    """
    repeats: AllowableTroolean = AllowableTroolean.FORBIDDEN
    """
    Whether this relationship supports repeated interactions.
    """
    attributes: dict[Name, ColumnSpec] = {}
    """
    Relationship attribute definitions.
    """

    @property
    def entity_class_names(self) -> list[str]:
        return list(self.entities.keys())


class ColumnSpec(BaseModel):
    layout: AttrLayout = AttrLayout.SCALAR
    """
    The attribute layout (whether and how multiple values are supported).
    """

    vector_size: int | None = None
    """
    The dimensionality of the vector, for sparse and vector columns.
    """



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from time import perf_counter


class RateLimit:
    """
    Class for rate-limiting performance updates.

    Args:
        hertz:
            The update frequency in Hertz.
    """

    hertz: float
    _interval: float

    _last_update: float = 0.0

    def __init__(self, hertz: float = 5):
        self.hertz = hertz
        self._interval = 1 / hertz

    def want_update(self, now: float | None = None) -> bool:
        """
        Check if we are ready for an update.
        """
        if now is None:
            now = perf_counter()

        if now - self._last_update >= self._interval:
            return True
        else:
            return False

    def mark_update(self, now: float | None = None):
        """
        Mark that the system has been updated.
        """
        if now is None:
            now = perf_counter()
        self._last_update = now



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from __future__ import annotations

from threading import RLock
from uuid import UUID

import structlog
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    ProgressColumn,
    SpinnerColumn,
    Task,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeRemainingColumn,
)
from rich.progress import Progress as ProgressImpl
from rich.text import Text
from typing_extensions import TYPE_CHECKING, override

from .._console import console, get_live
from .._proxy import get_logger
from ._base import Progress
from ._formats import field_format

if TYPE_CHECKING:
    from ..multiprocess._protocol import ProgressMessage

_log = get_logger("lenskit.logging.progress")
_pb_lock = RLock()
_progress: ProgressImpl | None = None
_active_bars: dict[UUID, RichProgress] = {}


class RichProgress(Progress):
    """
    Progress bark backed by Rich.
    """

    label: str
    logger: structlog.stdlib.BoundLogger
    _field_format: str | None = None
    _task: TaskID | None = None

    def __init__(
        self,
        label: str,
        total: int | float | None,
        fields: dict[str, str | None] | None,
        *,
        uuid: UUID | None = None,
    ):
        super().__init__(uuid=uuid)

        self.label = label
        self.total = total

        self.logger = _log.bind(label=label, uuid=str(self.uuid))

        self._task = self._install()

        if fields:
            self._field_format = _make_format(fields)

    @classmethod
    def handle_message(cls, update: ProgressMessage):
        with _pb_lock:
            pb = _active_bars.get(update.progress_id)
            if pb is None:
                pb = cls(update.label, update.total, {}, uuid=update.progress_id)

        if update.finished:
            pb.finish()
            return

        fields = {}
        if update.fields:
            cls._field_format = _make_format(
                {n: f.format for (n, f) in update.fields.items() if f.value is not None}
            )
            fields = {n: f.value for (n, f) in update.fields.items() if f.value is not None}
        else:
            cls._field_format = ""

        pb.update(advance=0, completed=update.completed, total=update.total, **fields)

    def update(
        self,
        advance: int = 1,
        completed: int | float | None = None,
        total: int | float | None = None,
        **kwargs: float | int | str,
    ):
        extra = ""
        if self._field_format:
            extra = self._field_format.format(**kwargs)
        if total is not None:
            self.total = total
        if _progress is not None:
            _progress.update(
                self._task,  # type: ignore
                advance=advance if completed is None else None,
                completed=completed,
                total=total,
                extra=extra,
            )

    def finish(self):
        self._remove()

    def _install(self) -> TaskID | None:
        global _progress
        self.logger.debug("installing progress bar")
        live = get_live()
        if live is None:
            return None

        with _pb_lock:
            if _progress is None:
                _progress = ProgressImpl(
                    SpinnerColumn(),
                    TextColumn("[progress.description]{task.description}"),
                    BarColumn(),
                    MofNCompleteColumn(),
                    RateColumn(),
                    TaskProgressColumn(),
                    TimeRemainingColumn(),
                    TextColumn("{task.fields[extra]}"),
                    console=console,
                )
                live.update(_progress)

            _active_bars[self.uuid] = self
            return _progress.add_task(self.label, total=self.total, extra="")

    def _remove(self):
        live = get_live()
        if live is None or _progress is None:
            return
        if self.uuid not in _active_bars:
            return
        if self._task is None:
            return

        self.logger.debug("uninstalling progress bar")

        with _pb_lock:
            _progress.remove_task(self._task)
            del _active_bars[self.uuid]


def _make_format(fields: dict[str, str | None]) -> str:
    return ", ".join(
        [f"[json.key]{name}[/json.key]: {field_format(name, fs)}" for (name, fs) in fields.items()]
    )


class RateColumn(ProgressColumn):
    def __init__(self):
        super().__init__()

    @override
    def render(self, task: Task):
        speed = task.finished_speed or task.speed
        if speed is None:
            disp = "?"
        elif speed > 1000:
            disp = "{:d} it/s".format(int(speed))
        elif speed > 1:
            disp = "{:.3g} it/s".format(speed)
        else:
            disp = "{:.3g} s/it".format(1.0 / speed)

        return Text(disp, "progress.percentage")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from uuid import UUID

import numpy as np
from typing_extensions import assert_type

from pytest import raises, warns

from lenskit.pipeline import PipelineBuilder, PipelineError
from lenskit.pipeline.nodes import InputNode, Node
from lenskit.pipeline.types import TypecheckWarning


def test_init_empty():
    pipe = PipelineBuilder()
    assert len(pipe.nodes()) == 0


def test_create_input():
    "create an input node"
    pipe = PipelineBuilder()
    src = pipe.create_input("user", int, str)
    assert_type(src, Node[int | str])
    assert isinstance(src, InputNode)
    assert src.name == "user"
    assert src.types == set([int, str])

    assert len(pipe.nodes()) == 1
    assert pipe.node("user") is src


def test_lookup_optional():
    "lookup a node without failing"
    pipe = PipelineBuilder()
    pipe.create_input("user", int, str)

    assert pipe.node("item", missing="none") is None


def test_input_no_component():
    pipe = PipelineBuilder()
    pipe.create_input("user", int, str)
    pipe = pipe.build()

    assert pipe.component("user") is None


def test_lookup_missing():
    "lookup a node without failing"
    pipe = PipelineBuilder()
    pipe.create_input("user", int, str)

    with raises(KeyError):
        pipe.node("item")


def test_dup_input_fails():
    "create an input node"
    pipe = PipelineBuilder()
    pipe.create_input("user", int, str)

    with raises(ValueError, match="has node"):
        pipe.create_input("user", UUID)


def test_dup_component_fails():
    "create an input node"
    pipe = PipelineBuilder()
    pipe.create_input("user", int, str)

    with raises(ValueError, match="has node"):
        pipe.add_component("user", lambda x: x)  # type: ignore


def test_dup_alias_fails():
    "create an input node"
    pipe = PipelineBuilder()
    n = pipe.create_input("user", int, str)

    with raises(ValueError, match="has node"):
        pipe.alias("user", n)  # type: ignore


def test_alias():
    "alias a node"
    pipe = PipelineBuilder()
    user = pipe.create_input("user", int, str)

    pipe.alias("person", user)

    assert pipe.node("person") is user

    # aliases conflict as well
    with raises(ValueError):
        pipe.create_input("person", bytes)


def test_component_type():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)

    def incr(msg: str) -> str:
        return msg

    node = pipe.add_component("return", incr, msg=msg)
    assert node.name == "return"
    assert node.types == set([str])


def test_single_input():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)

    def incr(msg: str) -> str:
        return msg

    node = pipe.add_component("return", incr, msg=msg)

    pipe = pipe.build()

    ret = pipe.run(node, msg="hello")
    assert ret == "hello"

    ret = pipe.run(node, msg="world")
    assert ret == "world"


def test_single_input_required():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)

    def incr(msg: str) -> str:
        return msg

    node = pipe.add_component("return", incr, msg=msg)

    pipe = pipe.build()
    with raises(PipelineError, match="not specified"):
        pipe.run(node)


def test_single_optional_input():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str, None)

    def fill(msg: str | None) -> str:
        return msg if msg is not None else "undefined"

    node = pipe.add_component("return", fill, msg=msg)

    pipe = pipe.build()
    assert pipe.run(node) == "undefined"
    assert pipe.run(node, msg="hello") == "hello"


def test_single_input_typecheck():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)

    def incr(msg: str) -> str:
        return msg

    node = pipe.add_component("return", incr, msg=msg)

    pipe = pipe.build()
    with raises(TypeError):
        pipe.run(node, msg=47)


def test_component_type_mismatch():
    pipe = PipelineBuilder()

    def incr(msg: str) -> str:
        return msg

    node = pipe.add_component("return", incr, msg=47)
    pipe = pipe.build()
    with raises(TypeError):
        pipe.run(node)


def test_component_unwired_input():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)

    def ident(msg: str, m2: str | None) -> str:
        if m2:
            return msg + m2
        else:
            return msg

    node = pipe.add_component("return", ident, msg=msg)
    pipe = pipe.build()
    assert pipe.run(node, msg="hello") == "hello"


def test_extract_component():
    pipe = PipelineBuilder()
    msg = pipe.create_input("msg", str)

    def incr(msg: str) -> str:
        return msg

    pipe.add_component("return", incr, msg=msg)
    pipe = pipe.build()

    assert pipe.component("return") is incr


def test_chain():
    pipe = PipelineBuilder()
    x = pipe.create_input("x", int)

    def incr(x: int) -> int:
        return x + 1

    def triple(x: int) -> int:
        return x * 3

    ni = pipe.add_component("incr", incr, x=x)
    nt = pipe.add_component("triple", triple, x=ni)
    pipe.default_component(nt)

    pipe = pipe.build()
    # run default pipe
    ret = pipe.run(x=1)
    assert ret == 6

    # run explicitly
    assert pipe.run(nt, x=2) == 9

    # run only first node
    assert pipe.run(ni, x=10) == 11


def test_chain_component_names():
    pipe = PipelineBuilder()
    x = pipe.create_input("x", int)

    def incr(x: int) -> int:
        return x + 1

    def triple(x: int) -> int:
        return x * 3

    ni = pipe.add_component("incr", incr, x=x)
    nt = pipe.add_component("triple", triple, x=ni)
    pipe.default_component(nt)

    pipe = pipe.build()
    assert pipe.component_names() == ["incr", "triple"]


def test_simple_graph():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    na = pipe.add_component("add", add, x=nd, y=b)
    pipe.default_component("add")

    pipe = pipe.build()
    assert pipe.run(a=1, b=7) == 9
    assert pipe.run(na, a=3, b=7) == 13
    assert pipe.run(nd, a=3, b=7) == 6


def test_cycle():
    pipe = PipelineBuilder()
    b = pipe.create_input("b", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double)
    na = pipe.add_component("add", add, x=nd, y=b)
    pipe.connect(nd, x=na)

    with raises(PipelineError, match="cycle"):
        pipe.build()


def test_replace_component():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def double(x: int) -> int:
        return x * 2

    def triple(x: int) -> int:
        return x * 3

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    na = pipe.add_component("add", add, x=nd, y=b)
    pipe.default_component(na)

    nt = pipe.replace_component("double", triple, x=a)

    pipe = pipe.build()

    # run through the end
    assert pipe.run(a=1, b=7) == 10
    assert pipe.run(na, a=3, b=7) == 16
    # run only the first component
    assert pipe.run(nt, a=3, b=7) == 9

    # old node should be missing!
    with raises(PipelineError, match="not in pipeline"):
        pipe.run(nd, a=3, b=7)


def test_default_wiring():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    pipe.default_connection("y", b)

    nd = pipe.add_component("double", double, x=a)
    na = pipe.add_component("add", add, x=nd)
    pipe.default_component(na)

    pipe = pipe.build()
    assert pipe.run(a=1, b=7) == 9
    assert pipe.run(na, a=3, b=7) == 13


def test_run_by_name():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    pipe.add_component("add", add, x=nd, y=b)

    pipe = pipe.build()
    assert pipe.run("double", a=1, b=7) == 2


def test_run_tuple_name():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    pipe.add_component("add", add, x=nd, y=b)

    pipe = pipe.build()
    res = pipe.run(("double",), a=1, b=7)
    assert isinstance(res, tuple)
    assert res[0] == 2


def test_run_tuple_pair():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    pipe.add_component("add", add, x=nd, y=b)

    pipe = pipe.build()
    res = pipe.run(("double", "add"), a=1, b=7)
    assert isinstance(res, tuple)
    d, a = res
    assert d == 2
    assert a == 9


def test_invalid_type():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    pipe.add_component("add", add, x=nd, y=b)

    pipe = pipe.build()
    with raises(TypeError):
        pipe.run("add", a=1, b="seven")


def test_run_by_alias():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    na = pipe.add_component("add", add, x=nd, y=b)

    pipe.alias("result", na)

    pipe = pipe.build()
    assert pipe.run("result", a=1, b=7) == 9


def test_run_all():
    pipe = PipelineBuilder("test", "7.2")
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    na = pipe.add_component("add", add, x=nd, y=b)

    pipe.alias("result", na)

    pipe = pipe.build()
    state = pipe.run_all(a=1, b=7)
    assert state["double"] == 2
    assert state["add"] == 9
    assert state["result"] == 9

    assert state.meta is not None
    assert state.meta.name == "test"
    assert state.meta.version == "7.2"
    assert state.meta.hash == pipe.config_hash


def test_run_all_limit():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    na = pipe.add_component("add", add, x=nd, y=b)

    pipe.alias("result", na)

    pipe = pipe.build()
    state = pipe.run_all("double", a=1, b=7)
    assert state["double"] == 2
    assert "add" not in state
    assert "result" not in state


def test_connect_literal():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    na = pipe.add_component("add", add, x=nd, y=2)

    pipe = pipe.build()
    assert pipe.run(na, a=3) == 8


def test_connect_literal_explicit():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    na = pipe.add_component("add", add, x=nd, y=pipe.literal(2))

    pipe = pipe.build()
    assert pipe.run(na, a=3) == 8


def test_fail_missing_input():
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)
    b = pipe.create_input("b", int)

    def double(x: int) -> int:
        return x * 2

    def add(x: int, y: int) -> int:
        return x + y

    nd = pipe.add_component("double", double, x=a)
    na = pipe.add_component("add", add, x=nd, y=b)

    pipe = pipe.build()

    with raises(PipelineError, match=r"input.*not specified"):
        pipe.run(na, a=3)

    # missing inputs only matter if they are required
    assert pipe.run(nd, a=3) == 6


def test_pipeline_component_default():
    """
    Test that the default component is run correctly.
    """
    pipe = PipelineBuilder()
    a = pipe.create_input("a", int)

    def add(x, y):  # type: ignore
        return x + y  # type: ignore

    with warns(TypecheckWarning):
        pipe.add_component("add", add, x=np.arange(10), y=a)  # type: ignore
    pipe.default_component("add")

    cfg = pipe.build_config()
    assert cfg.default == "add"

    pipe = pipe.build()
    # the component runs
    assert np.all(pipe.run("add", a=5) == np.arange(5, 15))

    # the component is the default
    assert np.all(pipe.run(a=5) == np.arange(5, 15))



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from __future__ import annotations

import warnings
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, Protocol, overload, runtime_checkable

from ._base import Progress

if TYPE_CHECKING:
    from ..multiprocess._protocol import ProgressMessage

_backend: ProgressBackend = Progress


@runtime_checkable
class ProgressBackend(Protocol):
    """
    Protocol implemented by progress backend objects.
    """

    def __call__(
        self, label: str, total: int | float | None, fields: Mapping[str, str | None] | None
    ) -> Progress: ...

    @classmethod
    def handle_message(cls, update: ProgressMessage): ...


@overload
def set_progress_impl(name: Literal["rich", "notebook", "none"] | None) -> None: ...
@overload
def set_progress_impl(backend: ProgressBackend, /) -> None: ...
def set_progress_impl(name: str | ProgressBackend | None, *options: Any) -> None:
    """
    Set the progress bar implementation.
    """
    global _backend

    match name:
        case "notebook":
            try:
                from ._notebook import JupyterProgress

                _backend = JupyterProgress
            except ImportError:
                warnings.warn("notebook progress backend needs ipywidgets")
                _backend = Progress

        case "rich":
            from ._rich import RichProgress

            _backend = RichProgress

        case "none" | None:
            _backend = Progress

        case ProgressBackend():
            _backend = name

        case _:
            raise ValueError(f"unknown progress backend {name}")


def item_progress(
    label: str, total: int | None = None, fields: Mapping[str, str | None] | None = None
) -> Progress:
    """
    Create a progress bar for distinct, counted items.

    Args:
        label:
            The progress bar label.
        total:
            The total number of items.
        fields:
            Additional fields to report with the progress bar (such as a current
            loss).  These are specified as a dictionary mapping field names to
            format strings (the pieces inside ``{...}`` in :meth:`str.format`),
            and the values come from extra kwargs to :meth:`Progress.update`;
            mapping to ``None`` use default ``str`` formatting.
    """
    return _backend(label, total, fields)


def progress_backend() -> ProgressBackend:
    return _backend



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"Definition of the component interfaces."

# pyright: strict
from __future__ import annotations

import json
import warnings
from abc import ABC, abstractmethod
from inspect import isabstract, signature
from types import FunctionType, NoneType

from pydantic import BaseModel, JsonValue, TypeAdapter
from typing_extensions import (
    Any,
    Callable,
    Generic,
    Mapping,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeVar,
    get_origin,
    get_type_hints,
    runtime_checkable,
)

from .types import Lazy, TypecheckWarning

P = ParamSpec("P")
T = TypeVar("T")
CFG = TypeVar("CFG")
CArgs = ParamSpec("CArgs", default=...)
"""
Argument type for a component.  It is difficult to actually specify this, but
using this default parameter spec allows :class:`Component` subclasses to
typecheck by declaring the base class :meth:`~Component.__call__` to have
unknown parameters.
"""
# COut is only return, so Component[U] can be assigned to Component[T] if U ≼ T.
COut = TypeVar("COut", covariant=True, default=Any)
"""
Return type for a component.
"""
PipelineFunction: TypeAlias = Callable[..., COut]
"""
Pure-function interface for pipeline functions.
"""


@runtime_checkable
class ComponentConstructor(Protocol, Generic[CFG, COut]):
    """
    Protocol for component constructors.
    """

    def __call__(self, config: CFG | None = None) -> Component[COut]: ...

    def config_class(self) -> type[CFG] | None: ...

    def validate_config(self, data: Any = None) -> CFG | None: ...


class Component(ABC, Generic[COut, CArgs]):
    """
    Base class for pipeline component objects.  Any component that is not just a
    function should extend this class.

    Pipeline components support configuration (e.g., hyperparameters or random
    seeds) through Pydantic models or Python dataclasses; see
    :ref:`component-config` for further details.  If the pipeline's
    configuration class is ``C``, it has the following:

    1. The configuration is exposed through an instance variable ``config``.
    2. The constructor accepts the configuration object as its first parameter,
       also named ``config``, and saves this in the member variable.

    The base class constructor handles both of these, so long as you declare the
    type of the ``config`` member::

        class MyComponent(Component):
            config: MyComponentConfig

            ...

    If you do not declare a ``config`` attribute, the base class will assume the
    pipeline uses no configuration.

    To work as components, derived classes also need to implement a ``__call__``
    method to perform their operations.

    Args:
        config:
            The configuration object.  If ``None``, the configuration class will
            be instantiated with ``kwargs``.

    Stability:
        Full
    """

    config: Any = None
    """
    The component configuration object.  Component classes that support
    configuration **must** redefine this attribute with their specific
    configuration class type, which can be a Python dataclass or a Pydantic
    model class.
    """

    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__(**kwargs)
        if not isabstract(cls):
            try:
                ct = cls.config_class(return_any=True)
            except NameError:
                # can't look up the config class, so we can't check it
                pass
            else:
                if ct == Any:
                    warnings.warn(
                        "component class {} does not define a config attribute type".format(
                            cls.__qualname__
                        ),
                        stacklevel=2,
                    )

    def __init__(self, config: object | None = None, **kwargs: Any):
        if config is None:
            config = self.validate_config(kwargs)
        elif kwargs:
            raise RuntimeError("cannot supply both a configuration object and kwargs")

        cfg_cls = self.config_class(return_any=True)
        if cfg_cls == Any:
            warnings.warn(
                "component class {} does not define a config attribute type".format(
                    self.__class__.__qualname__
                ),
                stacklevel=2,
            )
        elif cfg_cls and not isinstance(config, cfg_cls):
            raise TypeError(f"invalid configuration type {type(config)}")

        self.config = config

    @classmethod
    def config_class(cls, return_any: bool = False) -> type | None:
        hints = get_type_hints(cls)
        ct = hints.get("config", None)
        if ct == NoneType:
            return None
        elif ct is None or ct == Any:
            if return_any:
                return ct
            else:
                return None
        elif isinstance(ct, type):
            return ct
        else:
            warnings.warn("config attribute is not annotated with a plain type", stacklevel=2)
            return get_origin(ct)

    def dump_config(self) -> dict[str, JsonValue]:
        """
        Dump the configuration to JSON-serializable format.
        """
        cfg_cls = self.config_class()
        if cfg_cls:
            return TypeAdapter(cfg_cls).dump_python(self.config, mode="json")  # type: ignore
        else:
            return {}

    @classmethod
    def validate_config(cls, data: Mapping[str, JsonValue] | None = None) -> object | None:
        """
        Validate and return a configuration object for this component.
        """
        if data is None:
            data = {}
        cfg_cls = cls.config_class()
        if cfg_cls:
            return TypeAdapter(cfg_cls).validate_python(data)  # type: ignore
        elif data:  # pragma: nocover
            raise RuntimeError(
                "supplied configuration options but {} has no config class".format(cls.__name__)
            )
        else:
            return None

    @abstractmethod
    def __call__(self, *args: CArgs.args, **kwargs: CArgs.kwargs) -> COut:  # pragma: nocover
        """
        Run the pipeline's operation and produce a result.  This is the key
        method for components to implement.
        """
        ...

    def __repr__(self) -> str:
        params = json.dumps(self.dump_config(), indent=4)
        return f"<{self.__class__.__name__} {params}>"


class PlaceholderConfig(BaseModel, extra="allow"):
    """
    Configuration for the placeholder component.
    """

    pass


class Placeholder(Component[Any]):
    """
    Simple no-op component to use as a placeholder in partial pipelines.
    """

    config: PlaceholderConfig

    def __call__(self, **kwargs: Any) -> Any:
        raise NotImplementedError("attempted to invoke placeholder component")


def component_inputs(
    component: Component[COut] | ComponentConstructor[Any, COut] | PipelineFunction[COut],
    *,
    warn_on_missing: bool = True,
) -> dict[str, type | None]:
    if isinstance(component, FunctionType):
        function = component
    elif hasattr(component, "__call__"):
        function = getattr(component, "__call__")
    else:
        raise TypeError("invalid component " + repr(component))

    types = get_type_hints(function)
    sig = signature(function)

    inputs: dict[str, type | None] = {}
    for param in sig.parameters.values():
        if param.name == "self":
            continue

        if pt := types.get(param.name, None):
            inputs[param.name] = pt
        else:
            if warn_on_missing:
                warnings.warn(
                    f"parameter {param.name} of component {component} has no type annotation",
                    TypecheckWarning,
                    2,
                )
            inputs[param.name] = None

    return inputs


def component_return_type(
    component: Component[COut] | ComponentConstructor[Any, COut] | PipelineFunction[COut],
) -> type | None:
    if isinstance(component, FunctionType):
        function = component
    elif hasattr(component, "__call__"):
        function = getattr(component, "__call__")
    else:
        raise TypeError("invalid component " + repr(component))

    types = get_type_hints(function)
    return types.get("return", None)


def fallback_on_none(primary: T | None, fallback: Lazy[T]) -> T:
    """
    Fallback to a second component if the primary input is `None`.

    Stability:
        Caller
    """
    if primary is not None:
        return primary
    else:
        return fallback.get()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Allow LensKit to be run as a module.
"""

from lenskit.cli import main

main()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from pathlib import Path

import click

from lenskit.data.amazon import load_amazon_ratings
from lenskit.data.collection._list import ListILC
from lenskit.data.movielens import load_movielens
from lenskit.logging import get_logger

_log = get_logger(__name__)


@click.command("convert")
@click.option("--movielens", "format", flag_value="movielens", help="Convert MovieLens data.")
@click.option("--amazon", "format", flag_value="amazon", help="Convert Amazon rating data.")
@click.option(
    "--item-lists", is_flag=True, help="Convert to an ItemListCollection instead of Dataset."
)
@click.argument("src", type=Path, nargs=-1, required=True)
@click.argument("dst", type=Path, required=True)
def convert(format: str | None, src: list[Path], dst: Path, item_lists: bool = False):
    """
    Convert data into the LensKit native format.

    By default, this will create a LensKit Dataset.  With --item-lists, it will
    extract the default interactions into an ItemListCollection and save it in
    native format.
    """

    log = _log.bind(src=[str(p) for p in src])

    match format:
        case None:
            _log.error("no data format specified")
            raise click.UsageError("no data format specified")
        case "movielens":
            log.info("loading MovieLens data")
            if len(src) != 1:
                log.error("received %d source paths, MovieLens only takes one", len(src))
            data = load_movielens(src[0])
        case "amazon":
            data = load_amazon_ratings(*src)
        case _:
            raise ValueError(f"unknown data format {format}")

    log = _log.bind(dst=str(dst), name=data.name)
    if item_lists:
        icls = data.default_interaction_class()
        log.info("extracting %s data", icls)
        rels = data.interactions(icls).matrix()
        ilc: ListILC = ListILC.from_dict(
            {(rels.row_vocabulary.id(i),): rels.row_items(number=i) for i in range(rels.n_rows)},  # type: ignore
            f"{rels.row_type}_id",
        )
        log.info("saving to native Parquet format")
        ilc.save_parquet(dst)
    else:
        log.info("saving data in native format")
        data.save(dst)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Basic Top-*N* ranking.
"""

import logging

from pydantic import BaseModel

from lenskit.data import ItemList
from lenskit.pipeline.components import Component

_log = logging.getLogger(__name__)


class TopNConfig(BaseModel):
    """
    Configuration for top-N ranking.
    """

    n: int | None = None
    """
    The number of items to return. -1 or ``None`` to return all scored items.
    """


class TopNRanker(Component[ItemList]):
    """
    Rank scored items by their score and take the top *N*.  The ranking length
    can be passed either at runtime or at component instantiation time, with the
    latter taking precedence.

    Stability:
        Caller
    """

    config: TopNConfig
    "Configuration object."

    def __call__(self, *, items: ItemList, n: int | None = None) -> ItemList:
        """
        Rank the items.

        Args:
            items:
                The items to rank, with scores.  Items with missing scores are
                not included in the final ranking.
            n:
                The number of items to return, or -1 to return all scored items.
                If ``None``, the length configured at construction time is used.

        Returns:
            An ordered list of items, with scores and all other attributes
            preserved.
        """
        if n is None:
            n = self.config.n or -1

        if n >= 0:
            _log.debug("ranking top %d of %d items", n, len(items))
        else:
            _log.debug("ranking all of %d items", len(items))

        return items.top_n(n)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
PyTorch utility functions.
"""

import functools

import numpy as np
import torch


def inference_mode(func):
    """
    Function decorator that puts PyTorch in inference mode.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with torch.inference_mode():
            return func(*args, **kwargs)

    return wrapper


def sparse_row(mat: torch.Tensor, row: int) -> torch.Tensor:
    """
    Get a row of a sparse (CSR) tensor.  This is needed because indexing a
    tensor does not work in inference mode.
    """

    assert mat.is_sparse_csr

    cri = mat.crow_indices()
    sp = cri[row]
    ep = cri[row + 1]

    cs = mat.col_indices()
    vs = mat.values()
    return torch.sparse_coo_tensor(
        indices=cs[sp:ep].reshape(1, -1), values=vs[sp:ep], size=mat.shape[1:]
    )


def safe_tensor(array) -> torch.Tensor:
    """
    Safely convert an array into a NumPy tensor.  This includes copying it to
    writable memory if necessary.
    """
    if torch.is_tensor(array):
        return array

    arr = np.asarray(array)
    arr = np.require(arr, requirements="W")
    return torch.from_numpy(arr)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Flexible PyTorch matrix factorization models for LensKit.

The components in this package implement several matrix factorization models for
LensKit, and also serve as an example for practical PyTorch recommender
training.

.. stability:: internal

    FlexMF is provided as a preview release, and may change in the next months
    as we gain more experience with it.
"""

from ._base import FlexMFConfigBase, FlexMFScorerBase
from ._explicit import FlexMFExplicitConfig, FlexMFExplicitScorer
from ._implicit import FlexMFImplicitConfig, FlexMFImplicitScorer

__all__ = [
    "FlexMFConfigBase",
    "FlexMFScorerBase",
    "FlexMFExplicitConfig",
    "FlexMFExplicitScorer",
    "FlexMFImplicitConfig",
    "FlexMFImplicitScorer",
]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import logging
import pickle
from pathlib import Path

import numpy as np
import pandas as pd

from pytest import approx, fail, mark

from lenskit import recommend
from lenskit.data import (
    Dataset,
    ItemList,
    ItemListCollection,
    RecQuery,
    UserIDKey,
    from_interactions_df,
)
from lenskit.knn import UserKNNConfig, UserKNNScorer
from lenskit.metrics import call_metric, quick_measure_model
from lenskit.pipeline.common import predict_pipeline, topn_pipeline
from lenskit.testing import BasicComponentTests, ScorerTests
from lenskit.torch import inference_mode

_log = logging.getLogger(__name__)


class TestUserKNNExplicit(BasicComponentTests, ScorerTests):
    can_score = "some"
    component = UserKNNScorer
    expected_rmse = (0.855, 0.965)

    def test_internal_data(
        self, ml_ratings: pd.DataFrame, ml_ds: Dataset, trained_model: UserKNNScorer
    ):
        algo = trained_model
        # it should have computed correct means
        umeans = ml_ds.user_stats()["mean_rating"]
        mlmeans = pd.Series(algo.user_means, index=algo.users, name="mean")
        mlmeans.index.name = "user_id"
        umeans, mlmeans = umeans.align(mlmeans)
        assert mlmeans.values == approx(umeans.values)

        # we should be able to reconstruct rating values
        uir = ml_ratings.set_index(["user_id", "item_id"]).rating
        rates = algo.user_ratings.to_scipy().tocoo()
        ui_rbdf = pd.DataFrame(
            {
                "user_id": algo.users.ids(rates.row),
                "item_id": algo.items.ids(rates.col),
                "nrating": rates.data,
            }
        ).set_index(["user_id", "item_id"])
        ui_rbdf = ui_rbdf.join(mlmeans)
        ui_rbdf["rating"] = ui_rbdf["nrating"] + ui_rbdf["mean"]
        ui_rbdf["orig_rating"] = uir
        assert ui_rbdf.rating.values == approx(ui_rbdf.orig_rating.values)

        # running the predictor should work
        preds = algo(query=4, items=ItemList([1016]))
        assert len(preds) == 1
        assert preds.ids() == [1016]
        assert preds.scores() == approx([3.62221550680778])

    def verify_models_equivalent(self, orig: UserKNNScorer, copy: UserKNNScorer):
        assert np.all(copy.user_means == orig.user_means)


class TestUserKNNImplicit(BasicComponentTests, ScorerTests):
    can_score = "some"
    component = UserKNNScorer
    config = UserKNNConfig(max_nbrs=30, feedback="implicit")
    expected_ndcg = 0.03

    def verify_models_equivalent(self, orig, copy):
        assert copy.user_means is None
        assert copy.users == orig.users
        assert copy.items == orig.items


def test_uu_train(ml_ratings, ml_ds):
    algo = UserKNNScorer(k=30)
    algo.train(ml_ds)

    # we have data structures
    assert algo.user_means is not None
    assert algo.user_vectors is not None
    assert algo.user_ratings is not None

    # it should have computed correct means
    u_stats = ml_ds.user_stats()
    mlmeans = pd.Series(algo.user_means, index=algo.users.ids(), name="mean")
    mlmeans.index.name = "user_id"
    umeans, mlmeans = u_stats["mean_rating"].align(mlmeans)
    assert mlmeans.values == approx(umeans.values)

    # we should be able to reconstruct rating values
    uir = ml_ratings.set_index(["user_id", "item_id"]).rating
    rates = algo.user_ratings.to_scipy().tocoo()
    ui_rbdf = pd.DataFrame(
        {
            "user_id": algo.users.ids(rates.row),
            "item_id": algo.items.ids(rates.col),
            "nrating": rates.data,
        }
    ).set_index(["user_id", "item_id"])
    ui_rbdf = ui_rbdf.join(mlmeans)
    ui_rbdf["rating"] = ui_rbdf["nrating"] + ui_rbdf["mean"]
    ui_rbdf["orig_rating"] = uir
    assert ui_rbdf.rating.values == approx(ui_rbdf.orig_rating.values)


def test_uu_predict_one(ml_ds):
    algo = UserKNNScorer(k=30)
    algo.train(ml_ds)

    preds = algo(query=4, items=ItemList([1016]))
    assert len(preds) == 1
    assert preds.ids() == [1016]
    assert preds.scores() == approx([3.62221550680778])


def test_uu_predict_too_few(ml_ds):
    algo = UserKNNScorer(k=30, min_nbrs=2)
    algo.train(ml_ds)

    preds = algo(query=4, items=ItemList([2091]))
    assert len(preds) == 1
    assert preds.ids() == [2091]
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert all(preds.isna())


def test_uu_predict_too_few_blended(ml_ds):
    algo = UserKNNScorer(k=30, min_nbrs=2)
    algo.train(ml_ds)

    preds = algo(query=4, items=ItemList([1016, 2091]))
    assert len(preds) == 2
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert np.isnan(preds.loc[2091])
    assert preds.loc[1016] == approx(3.62221550680778)


def test_uu_predict_live_ratings(ml_ratings):
    algo = UserKNNScorer(k=30, min_nbrs=2)
    no4 = ml_ratings[ml_ratings.user_id != 4]
    no4 = from_interactions_df(no4)
    algo.train(no4)

    ratings = ItemList.from_df(ml_ratings[ml_ratings.user_id == 4][["item_id", "rating"]])

    query = RecQuery(20381, ratings)
    preds = algo(
        query=query,
        items=ItemList([1016, 2091]),
    )
    assert len(preds) == 2
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert np.isnan(preds.loc[2091])
    assert preds.loc[1016] == approx(3.62221550680778)


def test_uu_predict_unknown_empty(ml_ds):
    algo = UserKNNScorer(k=30, min_nbrs=2)
    algo.train(ml_ds)

    preds = algo(query=-28018, items=ItemList([1016, 2091]))
    assert len(preds) == 2
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert all(preds.isna())


@inference_mode
def test_uu_implicit(ml_ratings):
    "Train and use user-user on an implicit data set."
    algo = UserKNNScorer(k=20, feedback="implicit")
    data = ml_ratings.loc[:, ["user_id", "item_id"]]

    algo.train(from_interactions_df(data))
    assert algo.user_means is None

    mat = algo.user_vectors
    norms = np.linalg.norm(mat.todense(), axis=1)
    assert norms.shape == mat.shape[:1]
    assert np.allclose(norms, 1.0)

    preds = algo(query=50, items=ItemList([1, 2, 42]))
    preds = preds.scores("pandas", index="ids")
    assert preds is not None
    assert all(preds[preds.notna()] > 0)


@mark.slow
def test_uu_known_preds(ml_ds: Dataset):
    from lenskit import batch

    uknn = UserKNNScorer(k=30, min_sim=1.0e-6)
    pipe = predict_pipeline(uknn, fallback=False)
    _log.info("training %s on ml data", uknn)
    pipe.train(ml_ds)

    dir = Path(__file__).parent
    pred_file = dir / "user-user-preds.csv"
    _log.info("reading known predictions from %s", pred_file)
    known_preds = pd.read_csv(str(pred_file))
    _log.info("generating %d known predictions", len(known_preds))
    known = ItemListCollection.from_df(known_preds, UserIDKey)

    preds = batch.predict(pipe, known, n_jobs=1)
    preds = preds.to_df().drop(columns=["prediction"], errors="ignore")

    merged = pd.merge(known_preds.rename(columns={"prediction": "expected"}), preds)
    assert len(merged) == len(preds)
    merged["error"] = merged.expected - merged.score
    missing = merged.score.isna() & merged.expected.notna()
    if np.any(missing):
        bad = merged[merged.score.isna() & merged.expected.notna()]
        _log.error("%d missing predictions:\n%s", len(bad), bad)
        fail(f"missing predictions for {np.sum(missing)} items")

    err = merged.error
    err = err[err.notna()]
    if np.any(err.abs() > 0.01):
        bad = merged[merged.error.notna() & (merged.error.abs() > 0.01)]
        _log.error("%d erroneous predictions:\n%s", len(bad), bad)
        fail(f"{len(bad)} erroneous predictions")


@mark.slow
def test_uu_double_ratings(ml_ratings: pd.DataFrame):
    ml_ratings = ml_ratings.astype({"rating": "f8"})
    ds = from_interactions_df(ml_ratings)
    assert ds.interaction_matrix(format="pandas", field="rating")["rating"].dtype == np.float64
    model = UserKNNScorer(k=30, feedback="explicit")
    pipe = topn_pipeline(model)
    pipe.train(ds)

    # assert model.user_vectors_.dtype == torch.float32
    recs = recommend(pipe, 115, 10)
    assert len(recs) == 10



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: strict
from dataclasses import dataclass

from pytest import raises

from lenskit.diagnostics import PipelineError
from lenskit.pipeline import Component, PipelineBuilder
from lenskit.pipeline.nodes import ComponentInstanceNode


@dataclass
class PrefixConfig:
    prefix: str


class Prefixer(Component[str]):
    config: PrefixConfig

    def __call__(self, msg: str) -> str:
        return self.config.prefix + msg


def lowercase(msg: str) -> str:
    return msg.lower()


def test_modify_pipeline_nomod():
    builder = PipelineBuilder()
    msg = builder.create_input("msg", str)
    builder.add_component("prefix", Prefixer, PrefixConfig(prefix="scroll named "), msg=msg)
    builder.default_component("prefix")

    pipe = builder.build()
    assert pipe.run(msg="FOOBIE BLETCH") == "scroll named FOOBIE BLETCH"
    comp = pipe.node("prefix").component  # type: ignore

    b2 = pipe.modify()
    p2 = b2.build()
    n2 = p2.node("prefix")
    assert isinstance(n2, ComponentInstanceNode)
    assert isinstance(n2.component, Prefixer)
    assert n2.component is comp

    assert p2.run(msg="HACKEM MUCHE") == "scroll named HACKEM MUCHE"


def test_modify_pipeline_replace():
    builder = PipelineBuilder()
    msg = builder.create_input("msg", str)
    builder.add_component("prefix", Prefixer, PrefixConfig(prefix="scroll named "), msg=msg)
    builder.default_component("prefix")

    pipe = builder.build()
    assert pipe.run(msg="FOOBIE BLETCH") == "scroll named FOOBIE BLETCH"
    comp = pipe.node("prefix").component  # type: ignore

    b2 = pipe.modify()
    b2.replace_component("prefix", Prefixer, PrefixConfig(prefix="scroll called "))

    p2 = b2.build()
    n2 = p2.node("prefix")
    assert isinstance(n2, ComponentInstanceNode)
    assert isinstance(n2.component, Prefixer)
    assert n2.component is not comp

    assert p2.run(msg="HACKEM MUCHE") == "scroll called HACKEM MUCHE"


def test_modify_pipeline_new_input():
    builder = PipelineBuilder()
    msg = builder.create_input("msg", str)
    builder.add_component("prefix", Prefixer, PrefixConfig(prefix="scroll named "), msg=msg)
    builder.default_component("prefix")

    pipe = builder.build()
    assert pipe.run(msg="FOOBIE BLETCH") == "scroll named FOOBIE BLETCH"
    comp = pipe.node("prefix").component  # type: ignore

    b2 = pipe.modify()
    up = b2.add_component("upper", lowercase, msg=msg)
    b2.replace_component("prefix", Prefixer, PrefixConfig(prefix="scroll called "), msg=up)

    p2 = b2.build()
    n2 = p2.node("prefix")
    assert isinstance(n2, ComponentInstanceNode)
    assert isinstance(n2.component, Prefixer)
    assert n2.component is not comp

    assert p2.run(msg="HACKEM MUCHE") == "scroll called hackem muche"


def test_modify_pipeline_clear_input():
    builder = PipelineBuilder()
    msg = builder.create_input("msg", str)
    builder.add_component("prefix", Prefixer, PrefixConfig(prefix="scroll named "), msg=msg)
    builder.default_component("prefix")

    pipe = builder.build()
    assert pipe.run(msg="FOOBIE BLETCH") == "scroll named FOOBIE BLETCH"
    comp = pipe.node("prefix").component  # type: ignore

    b2 = pipe.modify()
    b2.replace_component("prefix", Prefixer, PrefixConfig(prefix="scroll called "))
    b2.clear_inputs("prefix")

    p2 = b2.build()
    n2 = p2.node("prefix")
    assert isinstance(n2, ComponentInstanceNode)
    assert isinstance(n2.component, Prefixer)
    assert n2.component is not comp

    with raises(TypeError):
        p2.run(msg="HACKEM MUCHE")



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Old home of the LensKit logging worker logic.

.. deprecated:: 2025.3
    This module is deprecated.  Import from :mod:`lenskit.logging.multiprocess`.
"""

from .multiprocess import WorkerContext, WorkerLogConfig, send_task

__all__ = ["WorkerContext", "WorkerLogConfig", "send_task"]



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from typing import TYPE_CHECKING

import numpy as np
from numpy.random import SeedSequence

from hypothesis import given
from hypothesis import strategies as st
from pytest import importorskip

from lenskit.random import random_generator

torch = importorskip("torch")

if TYPE_CHECKING:
    import torch


def test_torch_rng_none():
    rng = random_generator(type="torch")
    assert isinstance(rng, torch.Generator)


@given(st.integers(np.iinfo(np.int32).min, np.iinfo(np.int32).max))
def test_torch_rng_int(seed):
    rng = random_generator(seed, type="torch")
    assert isinstance(rng, torch.Generator)


@given(st.lists(st.integers(0)))
def test_torch_rng_seq(seed):
    rng = random_generator(seed, type="torch")
    assert isinstance(rng, torch.Generator)


@given(st.integers(0))
def test_torch_rng_seed(seed):
    seed = SeedSequence(seed)
    rng = random_generator(seed, type="torch")
    assert isinstance(rng, torch.Generator)


def test_torch_rng_from_numpy(rng):
    trng = random_generator(rng, type="torch")
    assert isinstance(trng, torch.Generator)


def test_torch_rng_from_numpy_bits(rng):
    trng = random_generator(rng.bit_generator, type="torch")
    assert isinstance(trng, torch.Generator)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import torch

import hypothesis.extra.numpy as nph
import hypothesis.strategies as st
from hypothesis import assume, given, settings

from lenskit.parallel.serialize import shm_deserialize, shm_serialize
from lenskit.testing import sparse_tensors


@given(st.lists(st.one_of(st.integers(), st.floats(allow_nan=False, allow_infinity=False))))
def test_basic_shm(obj):
    data = shm_serialize(obj)
    o2 = shm_deserialize(data)

    assert o2 == obj


@given(nph.arrays(nph.scalar_dtypes(), st.integers(0, 5000)))
def test_share_ndarray(arr):
    assume(np.all(np.isfinite(arr)))
    data = shm_serialize(arr)
    a2 = shm_deserialize(data)

    assert a2.shape == arr.shape
    assert np.all(a2 == arr)


@settings(deadline=5000)
@given(
    nph.arrays(
        st.one_of(nph.integer_dtypes(endianness="="), nph.floating_dtypes(endianness="=")),
        st.integers(0, 5000),
    )
)
def test_share_torch(arr):
    assume(np.all(np.isfinite(arr)))
    arr = torch.from_numpy(arr)
    data = shm_serialize(arr)
    a2 = shm_deserialize(data)

    assert a2.shape == arr.shape
    assert torch.all(a2 == arr)


@settings(deadline=1000)
@given(sparse_tensors(layout=["csr", "coo", "csc"]))
def test_share_sparse_torch(arr):
    data = shm_serialize(arr)
    a2 = shm_deserialize(data)

    assert a2.shape == arr.shape
    assert a2.layout == arr.layout
    assert torch.all(a2.to_dense() == arr.to_dense())



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

import numpy as np
import pandas as pd

from pytest import raises, warns

from lenskit.data.adapt import normalize_columns
from lenskit.data.types import AliasedColumn

SCROLLS = pd.DataFrame(
    {"appearance": ["FOOBIE BLETCH", "HACKEM MUCHE"], "type": ["identify", "teleport"]}
)


def test_normalize_column_name():
    df = normalize_columns(SCROLLS, "appearance")
    assert np.all(df.columns == ["appearance", "type"])


def test_normalize_column_from_index():
    df = normalize_columns(SCROLLS.set_index("appearance"), "appearance")
    assert np.all(df.columns == ["appearance", "type"])


def test_normalize_missing_column():
    with raises(KeyError):
        normalize_columns(SCROLLS, "bob")


def test_normalize_alias():
    col = AliasedColumn("random", ["appearance"])
    df = normalize_columns(SCROLLS, col)
    assert np.all(df.columns == ["random", "type"])


def test_normalize_alias_index():
    col = AliasedColumn("random", ["appearance"])
    df = normalize_columns(SCROLLS.set_index("appearance"), col)
    assert np.all(df.columns == ["random", "type"])


def test_normalize_missing_alias():
    col = AliasedColumn("random", ["color"])
    with raises(KeyError):
        normalize_columns(SCROLLS, col)


def test_normalize_alias_warn():
    col = AliasedColumn("random", ["appearance"], warn=True)
    with warns(DeprecationWarning):
        df = normalize_columns(SCROLLS, col)
    assert np.all(df.columns == ["random", "type"])


def test_normalize_alias_index_warn():
    col = AliasedColumn("random", ["appearance"], warn=True)
    with warns(DeprecationWarning):
        df = normalize_columns(SCROLLS.set_index("appearance"), col)
    assert np.all(df.columns == ["random", "type"])



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University.
# Copyright (C) 2023-2025 Drexel University.
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: basic
import numpy as np
import pandas as pd
import pyarrow as pa

from pytest import approx, mark, raises, skip, warns

from lenskit.data import Dataset, DatasetBuilder


def test_binarize_zero(ml_ds: Dataset) -> None:
    """Test that binarize removes items."""
    ratings_df = ml_ds.interactions().pandas()
    dsb = DatasetBuilder()
    ratings_df["user_num"] = ratings_df["user_num"].astype(int)
    ratings_df["item_num"] = ratings_df["item_num"].astype(int)
    ratings_df = ratings_df.rename(columns={"user_num": "user_id", "item_num": "item_id"})
    dsb.add_interactions(
        "rating", ratings_df, entities=["user", "item"], missing="insert", default=True
    )

    dsb.binarize_ratings("rating", min_pos_rating=3.5, method="zero")

    ds = dsb.build()
    ratings = ds.interactions().pandas()
    assert set(ratings["rating"]) <= {0.0, 1.0}
    assert all((ratings["rating"] == 1.0) == (ratings_df["rating"] >= 3.5))


def test_binarize_remove(ml_ds: Dataset) -> None:
    ratings_df = ml_ds.interactions().pandas()
    dsb = DatasetBuilder()
    ratings_df["user_num"] = ratings_df["user_num"].astype(int)
    ratings_df["item_num"] = ratings_df["item_num"].astype(int)
    ratings_df = ratings_df.rename(columns={"user_num": "user_id", "item_num": "item_id"})
    dsb.add_interactions(
        "rating", ratings_df, entities=["user", "item"], missing="insert", default=True
    )

    dsb.binarize_ratings("rating", min_pos_rating=3.5, method="remove")

    ds = dsb.build()
    ratings = ds.interactions().pandas()
    assert all(ratings["rating"] >= 3.5)


def test_binarize_error(ml_ds: Dataset) -> None:
    ratings_df = ml_ds.interactions().pandas()
    dsb = DatasetBuilder()
    ratings_df["user_num"] = ratings_df["user_num"].astype(int)
    ratings_df["item_num"] = ratings_df["item_num"].astype(int)
    ratings_df = ratings_df.rename(columns={"user_num": "user_id", "item_num": "item_id"})
    dsb.add_interactions(
        "rating", ratings_df, entities=["user", "item"], missing="insert", default=True
    )

    with raises(ValueError, match="method must be 'zero' or 'remove'"):
        dsb.binarize_ratings("rating", min_pos_rating=3.5, method="unknown")

    with raises(ValueError, match="min_pos_rating 6.0 is not in the range of ratings \[.*\]"):
        dsb.binarize_ratings("rating", min_pos_rating=6.0, method="zero")

    with raises(ValueError, match="min_pos_rating -1.0 is not in the range of ratings \[.*\]"):
        dsb.binarize_ratings("rating", min_pos_rating=-1.0, method="zero")
